package se.ericsson.eto.norarc.javaframe;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import jfdebug.JFClass;
import jfdebug.JFEvent;
import jfdebug.JFPath;
import se.ericsson.eto.norarc.javaframe.debug.Connection;
import se.ericsson.eto.norarc.javaframe.debug.DebugWindow;
import se.ericsson.eto.norarc.javaframe.debug.QualifiedNames;
import se.ericsson.eto.norarc.javaframe.debug.StateMachineClassData;

public class DebugControl {
	
	// the event currently happening and being logged
	private Event curEvent;
	// list of statemachines created during curEvent
	private List<StateMachine> created;
	// list with logged events
	private LinkedList<Event> past;
	// list with logged events that have been undone (empty unless the system is paused)
	private LinkedList<Event> future;
	// past never contains more events than maxEvents
	private int maxEvents;
	// a list of external signals not yet handled by state machines
	private Collection<Message> externalSigs;
	
	private Connection con;
	
	private StateMachineData selectedSM;
	
	// contains data for the different StateMachine classes/types
	public Map<Class, StateMachineClassData> smClassData;
	// used to get the UML qualified name for a java class and vice versa
	public QualifiedNames qNames;

	// true if JavaFrame is paused
	private boolean paused;
	public boolean isPaused() {return paused;}
	
	// the DebugScheduler's thread (internal signals are sent from this)
	private Thread debugThread;
	// the DebugScheduler
	private DebugScheduler sched;
	// sets the DebugScheduler and debugThread
	// should be called from the DebugScheduler's run() method
	protected void setDebugScheduler(DebugScheduler sched) {
		this.sched = sched;
		debugThread = Thread.currentThread();
	}

	private DebugWindow dWin;
	public void setDebugWindow(DebugWindow dWin) {
		this.dWin = dWin;
	}
	
	public DebugScheduler debugScheduler() {return sched;}
	
	// singleton object
	private static DebugControl instance = new DebugControl();
	public static DebugControl getInstance() {return instance;}

	
	// contains data for the existing StateMachines
	private Map<Long, StateMachineData> fsmData;
	
	protected void reset() {
		externalSigs = new LinkedList<Message>();
		created = new LinkedList<StateMachine>();
		debugThread = null;
		fsmData = new HashMap<Long, StateMachineData>();
		smClassData = new HashMap<Class, StateMachineClassData>();
		past = new LinkedList<Event>();
		future = new LinkedList<Event>();
		maxEvents = 100;
		paused = false;
	}
	private DebugControl() {
		reset();
	}
	
	/**
	 * Called before a transition
	 * @param fsm
	 * @param sig
	 */
	public void startEvent(StateMachine fsm, Message sig) {
		boolean external = externalSigs.remove(sig);
		StateMachineData data = fsmData.get(fsm.FSMid);
		if (data == null) {
			data = new StateMachineData(fsm);
			fsmData.put(fsm.FSMid, data);
		}
		curEvent = new Event(data, sig, external);
	}
	
	
	/**
	 * Called when a DebugMailbox recieves a signal
	 * @param mbox
	 * @param sig
	 * @return returns false if the mailbox should discard the signal
	 */
	public boolean signalRecieved(Message sig, MailBox mbox) {
		if (Thread.currentThread() == debugThread)
			if (curEvent != null) curEvent.addSignal(sig, mbox);
			else out("Ignoring signal " + sig + "(no event)");
		else {
			//TODO: external signal
			externalSigs.add(sig);
		}
		
		// TODO: returnvalue
		return true;
	}
	
	/**
	 * Called when a statemachine has terminated
	 * @param sm the terminated statemachine 
	 */
	public void eventTerminate(StateMachine sm) {
		if (curEvent != null) {
			if (curEvent.fsmData.id == sm.FSMid)
				curEvent.terminate();
			else out("Ignoring termination(" + sm.FSMid + ") (curEvent("
					+ curEvent.fsmData.id + "))");
		} else out("Ignoring termination(" + sm.FSMid + ") (no curEvent)");
	}
	
	
	/**
	 * Called when a statemachine has been created
	 * @param sm the created statemachine
	 */
	public void eventCreate(StateMachine sm) {
		if (curEvent != null)
			created.add(sm);
		else out("Ignoring creation(" + sm.FSMid + ") (no curEvent)");
	}
	/**
	 * finalizes curEvent and puts it in the past list
	 */
	public void stopEvent() {
		if (created.size() == 0)
			curEvent.finalize(null);
		else {
			// make an array with StateMachineData for the created statemachines
			StateMachineData smData[] = new StateMachineData[created.size()];
			int i = 0;
			for (StateMachine sm : created) {
				assert(!fsmData.containsKey(sm.FSMid));
				
				sm.messageBox = new DebugMailBox(sm.messageBox);
				sm.saveQueue = new DebugMailBox(sm.saveQueue);
				
				StateMachineData smd = new StateMachineData(sm);
				fsmData.put(smd.id, smd);
				smData[i] = smd;
				i++;
			}
			curEvent.finalize(smData);
			created.clear();
		}
		System.out.println(curEvent);
		
		past.addFirst(curEvent);
		curEvent = null;
		if (maxEvents < past.size()) {
			// remove past's oldest event
			Event e = past.removeLast();
			if (e.nextEvent != null) e.nextEvent.prevEvent = null;
			if (e.fsmData.prevEvent == e) e.fsmData.prevEvent = null;
			if (e.isTermination()) {
				// termination of e's statemachine can no longer be undone
				// remove e's StateMachineData
				assert(fsmData.containsKey(e.fsmData.id));
				out("Removing StateMachine " + e.fsmData.id);
				fsmData.remove(e.fsmData.id);

			}
			
		}
		if (qNames != null) sendEventToPlugin();
		if (dWin != null) dWin.setEvent(past.getFirst());
	}
	
	/**
	 * Undoes the most recent event in past
	 * Ignores termination/creation as this is handled when resuming execution
	 */
	synchronized public void previous() {
		if (past.size() == 0) return;
		synchronized (sched) {
			if (!sched.isPaused()) return;
			Event e = past.removeFirst();
			System.out.println("  Undoing: " + e);
			e.undo();
			future.addFirst(e);
			sched.readyFSM(e.fsmData.fsm);
			sendEventToPlugin();
			if (past.size() != 0) dWin.setEvent(past.getFirst());
			else dWin.setEvent(null);
		}
		return;
	}
	
	
	/**
	 * Redoes the first event in future
	 * Ignores termination/creation as this is handled when resuming execution
	 */
	synchronized public void next() {
		if (future.size() == 0) return;
		synchronized (sched) {
			if (!sched.isPaused()) return;
			Event e = future.removeFirst();
			System.out.println("  Redoing: " + e);
			e.redo();
			past.addFirst(e);
			sendEventToPlugin();
			dWin.setEvent(past.getFirst());
			
		}
		return;
	}
	public boolean emptyPast() {
		return (past.size() == 0);
	}
	public boolean emptyFuture() {
		return (future.size() == 0);
	}
	
	/**
	 * Pauses JavaFrame. Blocks until the DebugScheduler is paused
	 */
	synchronized public void pause() {
		synchronized (sched) {
			paused = true;
			sched.notify();
			while (!sched.isPaused()) {
				try {
					sched.wait();
				} catch (InterruptedException ie) {
					// the thread has been interrupted by Thread.interrupt()
					System.err.println("TRC Info("
							+ Thread.currentThread().toString() + "interrupted" );
				}
			}
		}
		out("Pausing");
	}
	
	/**
	 * Unpauses JavaFrame
	 */
	synchronized public void resume() {
		synchronized (sched) {
			paused = false;
			sched.notify();
			// the scheduler will resume once out of the synchronized block
			
			// empty the future-list
			while (future.size() != 0) {
				Event e = future.removeFirst();
				if (e.prevEvent != null) e.prevEvent.nextEvent = null;
				if (e.isTermination()) {
					// undo termination
					e.fsmData.addToMediators();
					StateMachine fsm = e.fsmData.fsm;
					fsm.owner.activeObjectList.add(fsm);
					fsm.moveStateMachine(sched);
				}
				StateMachineData smData[] = e.getCreated();
				if (smData == null) continue;
				for (StateMachineData smd : smData) {
					// undo creation
					out("Removing StateMachine " + smd.id);
					if (fsmData.remove(smd.id) == null)
						out("StateMachine not found");
					smd.fsm.moveStateMachine(null);
					smd.fsm.owner.removeActiveObject(smd.fsm);
				}
			}
		}
		out("Resuming execution");
	}
	void out(Object o) {
		System.out.println("-- DebugControl: " + o);
	}
	synchronized public void disConnected() {
		con = null;
	}
	synchronized public void connected(Connection con) {
		boolean b = !paused;
		if (b)	pause();
		this.con = con;
		try {
			qNames = new QualifiedNames();
			List<JFPath> pathList = (List<JFPath>) con.readObject();
			List<JFClass> classList = (List<JFClass>) con.readObject();
			System.out.println(pathList);
			System.out.println(classList);
			for (JFClass jfc : classList) {
				Class cls = null;
				try {cls = Class.forName(jfc.className);} catch (Exception e) {
					//e.printStackTrace();
				}
				if (cls == null) {
					out("Unable to load class: " + jfc.className);
					continue;
				}
				qNames.put(cls, jfc.qualifiedName);
				if (StateMachine.class.isAssignableFrom(cls)) {
					out(jfc.className + " is a StateMachine");
					StateMachineClassData smd = smClassData.get(cls);
					if (smd == null) {
						out("Creating new StateMachineClassData");
						smd = new StateMachineClassData(cls);
						smClassData.put(cls, smd);
					}
					smd.setQualifiedName(jfc.qualifiedName);
				} else if (Composite.class.isAssignableFrom(cls)) {
					out(jfc.className + " is a Composite");
					
				} else if (CompositeState.class.isAssignableFrom(cls)) {
					out(jfc.className + " is a CompositeState");
				} else out(jfc.className + " is not a Composite or a StateMachine");
			}
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		if (b) resume();
	}
	
	private void sendEventToPlugin() {
		if (con == null || qNames == null || past.size() == 0) return;
		
		Event event = past.getFirst();
		if (event == null) return;
		String aName = null;
		String bName = null;
		StateMachineData smd = event.fsmData;
		//TODO: check smd against selected StateMachine
		try {
			
			State state = event.beforeState;
			if (state != null && state.enclosingState != null) {
				Class cls = state.enclosingState.getClass();
				bName = qNames.getUmlName(cls);
				bName += "::" + state.stateName;
			}
			state = event.afterState;
			if (state != null && state.enclosingState != null) {
				Class cls = state.enclosingState.getClass();
				aName = qNames.getUmlName(cls);
				aName += "::" + state.stateName;
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		String sigName = qNames.getUmlName(event.sig.getClass());
		if (sigName == null) {
			if (event.sig instanceof TimerMsg)
				sigName = ":::timer";
			else {
				sigName = event.sig.getClass().getName();
				if (sigName.contains("."))
					sigName = sigName.substring(sigName.lastIndexOf(".") + 1);
			}
		}
		String smName = qNames.getUmlName(event.fsmData.fsm.getClass());
		
		System.out.println(bName + " -> " + aName);
		con.writeObject(new JFEvent(smName, bName, aName, sigName));
	}

}
