package se.ericsson.eto.norarc.javaframe;
import java.lang.reflect.Field;
import java.util.LinkedList;
import java.util.List;

public class Event {
	
	// list of statemachines created during the transition
	private StateMachineData[] created;
	public StateMachineData[] getCreated() {return created;}
	
	boolean externalSig;
	
	public boolean isInitial() {return (sig instanceof StartMessage);}
	// true if the transition ends in a final state
	private boolean termination;
	public boolean isTermination() {return termination;}
	// the signal that triggered the transition
	protected Message sig;
	// the event's statemachine
	protected StateMachineData fsmData;
	public StateMachineData getFsmData() {return fsmData;}
	// the statemachine's state before and the transition
	protected State beforeState;
	public State getBeforeState() {return beforeState;}
	protected State afterState;
	public State getAfterState() {return afterState;}
	// the statemachine's attributes before and after the transition
	protected Object[] beforeAttr;
	public Object[] getBeforeAttr() {return beforeAttr;}
	protected Object[] afterAttr;
	public Object[] getAfterAttr() {return afterAttr;}
	// a list of sent signals and recieving mailboxes
	protected List<MailBoxSignalTuple> sentSignals;
	
	// the next and previous events for the same statemachine
	protected Event prevEvent;
	protected Event nextEvent;
	
	class MailBoxSignalTuple {
		MailBox mbox;
		Message sig;
		
		MailBoxSignalTuple(MailBox mbox, Message sig) {
			this.mbox = mbox;
			this.sig = sig;
		}
	}
	
	/**
	 * @param fsmData the statemachine making the transition
	 * @param sig the signal triggering the transition
	 * @param external true if the signal is an external signal (timer, SMS, ...)
	 */
	public Event(StateMachineData fsmData, Message sig, boolean external) {
		externalSig = external;
		termination = false;
		sentSignals =  null;
		this.fsmData = fsmData;
		this.sig = sig;
		prevEvent = fsmData.prevEvent;
		if (prevEvent == null) {
			// no previous event for the statemachine
			// copy the statemachine's attributes
			beforeAttr = fsmData.copyFsm();
			beforeState = fsmData.fsm.currentState;
		} else if (fsmData.dirty) {
			// the statemachine's and prevEvent's attributes point to the same data
			// make a new copy of the attributes
			beforeAttr = fsmData.copyFsm();
			prevEvent.afterAttr = beforeAttr;
			beforeState = prevEvent.afterState;
		} else {
			// point to the attribute/state of the statemachine's previous event
			prevEvent.nextEvent = this;
			beforeAttr = fsmData.prevEvent.afterAttr;
			beforeState = fsmData.prevEvent.afterState;
		}
		fsmData.dirty = false;
	}
	
	/**
	 * Makes the event a termination event. For transitions ending in final state
	 */
	public void terminate() {
		termination = true;
	}
	
	/**
	 * Called when the transition has been made
	 * @param created a list of statemachine's created during the transition
	 */
	public void finalize(StateMachineData[] created) {
		this.created = created;
		afterAttr = fsmData.copyFsm();
		afterState = fsmData.fsm.currentState;
		fsmData.prevEvent = this;
	}
	
	/**
	 * Adds a sent signal to the current event
	 * @param sig the signal
	 * @param mbox the recieving mailbox
	 */
	public void addSignal(Message sig, MailBox mbox) {
		if (sentSignals == null) sentSignals = new LinkedList<MailBoxSignalTuple>();
		sentSignals.add(new MailBoxSignalTuple(mbox, sig));
	}

	
	/**
	 * Undoes the event's transition by setting the statemachine's data to the
	 * beforeAttr and beforeState data
	 */
	public void undo() {
		if (sentSignals != null)
			for (MailBoxSignalTuple m : sentSignals) {
				m.mbox.remove(m.sig);
			}
		if (!externalSig) fsmData.fsm.messageBox.addMessageFirst(sig);
		
		try {
			fsmData.setFsm(beforeAttr, beforeState);
		} catch (IllegalAccessException e) {
			throw new Error(e);
		}
		fsmData.prevEvent = prevEvent;
		fsmData.dirty = true;
	}
	
	/**
	 * Undoes the event's transition by setting the statemachine's data to the
	 * afterAttr and afterState data
	 */
	public void redo() {
		if (!externalSig) fsmData.fsm.messageBox.removeFirst();
		if (sentSignals != null)
			for (MailBoxSignalTuple m : sentSignals) {
				m.mbox.addMessage(m.sig);
			}
		try {
			fsmData.setFsm(afterAttr, afterState);
		} catch (IllegalAccessException e) {
			throw new Error(e);
		}
		fsmData.prevEvent = this;
		fsmData.dirty = true;
	}
	
	
	/**
	 * Helper function, gives more information than sig.toString()
	 * @param sig the signal
	 * @return a string representing the signal
	 */
	public static String sigToString(Message sig) {
		String c = sig.getClass().getName();
		c = c.substring(c.lastIndexOf('.') + 1);
		String s = c + "(";
		Field[] fields = sig.getClass().getDeclaredFields();
		boolean b = false;
		for (Field f : fields) {
			try {
				s += f.get(sig) + ", ";
				b = true;
			} catch (IllegalAccessException e) {}
		}

		if (b)s = s.substring(0, s.length()-2);
		s += ")";
		return s;
	}
	
	public String toString() {
		String c = fsmData.fsm.getClass().getName();
		c = c.substring(c.lastIndexOf('.') + 1);
		String s = "Event(" + c + "(";
		s += (fsmData.id) + ") ";
		s += sigToString(sig) + " State(";
		if (beforeState != null) s += beforeState.stateName();
		else s += "?";
		s += " -> ";
		if (afterState != null) s += afterState.stateName();
		else s += "?";
		s += "))";
		/*
		s += "\n Attributes(";
		Field fields[] = fsmData.classData.getAttributes();
		
		for (int i = 0; i < fields.length; i++) {
			Field f = fields[i];
			boolean b = !(f.isAccessible());
			if (b) f.setAccessible(true);
			
			s += "\n  " + f.getName() + "("
				+ beforeAttr[i] + " -> " + afterAttr[i] + ")";
			
			if (b) f.setAccessible(false);
		}
		s += ")\n SentSignals(";
		for (MailBoxSignalTuple t : sentSignals) {
			s += "\n  " + sigToString(t.sig);
		}
		s += ")\n)";*/
		if (externalSig) s += " (external)";
		if (isInitial()) s += " (start)";
		if (termination) s += " (terminated)";
		if (created != null) {
			s += " created(";
			for (StateMachineData smd : created) {
				s += " (" + smd.id + ") ";
			}
			s += ")";
		}
		return s;
	}

}
