package se.ericsson.eto.norarc.javaframe;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import se.ericsson.eto.norarc.javaframe.debug.StateMachineClassData;



/**
 * @author Jonas
 *
 */
public class StateMachineData {

	long id;
	protected StateMachine fsm;
	protected Event prevEvent;
	protected StateMachineClassData classData;
	public StateMachineClassData getClassData() {return classData;}
	public String getName() {
		String s = fsm.getClass().getName();
		s = s.substring(s.lastIndexOf('.') + 1) + "(" + id + ")";
		return s;
	}
	
	// if true the attributes of the statemachine should be copied back to prevEvent
	// before the next transition
	protected boolean dirty;
	
	/**
	 * @author Jonas
	 * 
	 */
	class MediatorData {
		Mediator fsmMed;
		Mediator fromMed;
		Object id;
		MediatorData(Mediator fsmMed, Mediator fromMed) {
			this.fsmMed = fsmMed;
			this.fromMed = fromMed;
		}
		MediatorData(Mediator fsmMed, SimpleIdRouterMediator fromMed, Object id) {
			this.fsmMed = fsmMed;
			this.fromMed = fromMed;
			this.id = id;
		}
	}
	
	MediatorData mediators[];
	
	/**
	 * Finds all mediators with one of fsm's mediators as address 
	 */
	private void collectAddresses() {
		try {
			List<MediatorData> medList = new LinkedList<MediatorData>();
			
			// get all the StateMachine's mediators
			Class mClass = Mediator.class;
			// TODO: only add mediators with fsm as address?
			List<Mediator> fsmAddresses = new ArrayList<Mediator>();

			for (Field f : classData.getMediators()) {
				boolean b = f.isAccessible();
				if (!b) f.setAccessible(true);
				Mediator fsmMed = (Mediator)f.get(fsm);
				if (!b) f.setAccessible(false);
				
				//if (fsmMed.address == fsm)
					fsmAddresses.add(fsmMed);
			}
			
			// Get all the statemachiene's owner's mediators
			for (Field f : fsm.owner.getClass().getDeclaredFields()) {
				if (mClass.isAssignableFrom(f.getType())) {
					
					boolean b = f.isAccessible();
					if (!b) f.setAccessible(true);
					Mediator fromMed = (Mediator)f.get(fsm.owner);
					if (!b) f.setAccessible(false);
					
					if (fromMed instanceof SimpleRouterMediator) {
						SimpleRouterMediator rm = (SimpleRouterMediator) fromMed;
						int j = 0;
						for (Object o : rm.mediatorList) {
							for (Mediator fsmMed : fsmAddresses) {
								if (o == fsmMed) {
									if (fromMed instanceof SimpleIdRouterMediator) {
										SimpleIdRouterMediator irm = (SimpleIdRouterMediator) rm;
										Object medId = irm.idList.get(j);
										medList.add(new MediatorData(fsmMed, irm, medId));
									} else medList.add(new MediatorData(fsmMed, fromMed));
								}
							}
							j++;
						}
					} else if (fromMed instanceof MultiCastMediator) {
						MultiCastMediator mcm = (MultiCastMediator) fromMed;
						for (Object o : mcm.mediatorList) {
							for (Mediator fsmMed : fsmAddresses) {
								if (o == fsmMed) {
									medList.add(new MediatorData(fsmMed, fromMed));
								}
							}
						}
						
					} else  {
						for (Mediator fsmMed : fsmAddresses) {
							if (fromMed.address == fsmMed) {
								medList.add(new MediatorData(fsmMed, fromMed));
							}
						}
					}
				}
			}
			mediators = medList.toArray(new MediatorData[0]);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
			throw new Error(e);
		}
	}
	
	/**
	 * Removes the statemachine's mediators from the addresses of other mediators
	 */
	protected void removeFromMediators() {
		for (MediatorData md : mediators) {
			md.fromMed.removeAddress(md.fsmMed);
		}
	}
	/**
	 * Adds the statemachine's mediators from the addresses other mediators
	 */
	protected void addToMediators() {
		for (MediatorData md : mediators) {
			if (md.fromMed instanceof SimpleIdRouterMediator) {
				SimpleIdRouterMediator idm = (SimpleIdRouterMediator) md.fromMed;
				idm.addId(md.id);
			}
			md.fromMed.addAddress(md.fsmMed);
		}
	}
	
	/**
	 * 
	 * @param fsm the statemachine
	 */
	StateMachineData(StateMachine fsm) {
		this.id = fsm.FSMid;
		this.fsm = fsm;
		dirty = false;
		
		DebugControl dc = DebugControl.getInstance();
		classData = dc.smClassData.get(fsm.getClass());
		if (classData == null) {
			classData = new StateMachineClassData(fsm.getClass());
			dc.smClassData.put(fsm.getClass(), classData);
		}
		collectAddresses();
	}
	
	
	/**
	 * Tries to copy an object by invoking its clone() method, if the object's
	 * class implements Cloneable and has an accessible clone() method
	 * @param o the object to copy
	 * @return the copied object (or a pointer to o, if cloning failed)
	 */
	private Object copyObject(Object o) {
		if (o instanceof Cloneable) {
			try {
				Method m = o.getClass().getMethod("clone");
				if (m.getReturnType() != Void.TYPE) {
					Object ret = m.invoke(o);
					return ret;
				}
				
			} catch (NoSuchMethodException e) {
			} catch (IllegalAccessException e) {
			} catch (InvocationTargetException e) {}
		}
		// if the method hasn't returned yet, invoking clone() failed
		// return the original object
		return o;
	}
	
	/**
	 * Copies the attributes of a statemachine to an array
	 * @param fsm the StateMachine to cloned
	 * @return an array with the attribute-values of fsm
	 */
	protected Object[] copyFsm() {
		
		try {
			//System.out.println("copying " + fsm);
			
			Object[] values = new Object[classData.getAttributes().length];
			Field[] fields = classData.getAttributes();
			for (int i = 0; i < fields.length; i++) {
				
				Field f = fields[i];
				//System.out.println("-- " + f.getName() + ":");
				Object o = null;
				
				// get the value of the field in fsm
				boolean b = f.isAccessible();
				if (!b) f.setAccessible(true);
				o = f.get(fsm);
				if (!b) f.setAccessible(false);
				
				if (o == null) {
					values[i] = null;
					continue;
				}
				if (f.getType().isPrimitive() || o instanceof String
						|| o instanceof  java.math.BigDecimal
						|| o instanceof java.math.BigInteger
						|| o instanceof Byte || o instanceof Double
						|| o instanceof Float || o instanceof Integer
						|| o instanceof Long || o instanceof Short
						|| o instanceof Character || o instanceof Boolean) {
					// immutable object; just copy the pointer
					// TODO: wrapper classes? ...
					values[i] = o;
					continue;
				}
				if (o instanceof Message) {
					try {
						values[i] = ((Message)o).duplicate();
						continue;
					} catch (CloneNotSupportedException e) {
						//System.out.print("Duplicate failed.");
						// no continue; will be handled by next if-test instead
					}
				}
				
				if (!(o instanceof Serializable)) {
					values[i] = copyObject(o);
				} else {
					// write the object to and read it from a byte array to flatten and copy it
					ByteArrayOutputStream bstream = new ByteArrayOutputStream();
					ObjectOutputStream out = new ObjectOutputStream(bstream);
					
					try {
						out.writeObject(o);
						ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(bstream.toByteArray()));
						values[i] = in.readObject();
						//System.out.println("Object serialized.");
						in.close();
					} catch (NotSerializableException e) {
						//System.out.print("Couldn't serialize field.");
						// serialization failed. try to clone it instead
						values[i] = copyObject(o);
					}
					out.close();
					bstream.close();
				}
			}
			
			//System.out.println("copied " + fsm);
			return values;
			
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return null;
	}
	
	/**
	 * sets the state and attributes of fsm
	 * @param attributes
	 * @param state
	 * @throws IllegalAccessException
	 */
	protected void setFsm(Object[] attributes, State state) throws IllegalAccessException {
		fsm.currentState = state;
		
		Field f[] = classData.getAttributes();
		for (int i = 0; i < f.length; i++) {
			boolean b = f[i].isAccessible();
			if (!b) f[i].setAccessible(true);
			f[i].set(fsm, attributes[i]);
			if (!b) f[i].setAccessible(false);
		}
	}
	
}
